<?php
/**
 * Project: 4SEO
 *
 * @author           Yannick Gaultier - Weeblr llc
 * @copyright        Copyright Weeblr llc - 2020-2023
 * @package          4SEO
 * @license          GNU General Public License version 3; see LICENSE.md
 * @version          4.5.1.2042
 * @date        2023-02-15
 */

use Joomla\CMS\Plugin;
use Weeblr\Wblib\Forseo\Wb;
use Weeblr\Wblib\Forseo\Factory as wblFactory;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;

defined('_JEXEC') or die;

/**
 * Handle commercial extension update authorization
 *
 * @since       2.5
 */
class PlgInstallerForseo extends Plugin\CMSPlugin
{
	/**
	 * @var    String  base update url, to decide whether to process the event or not
	 * @since  2.5
	 */
	private $baseUrl = [
		'https://u1.weeblr.com/dist/forseo/full' => 'forseo',
	];

	/**
	 * @var    String  extension identifier, to retrieve its params
	 * @since  2.5
	 */
	private $extension = 'pkg_forseo';

	/**
	 * @var    String    An id for your product, to be used by the web site when deciding to allow access
	 *                    Not mandatory, depends on subscription management system
	 * @since  2.5
	 */
	private $productId = '';

	/**
	 * @var    String    An edition type (full, free, lite,...) for the product
	 *                    Not mandatory, depends on subscription management system
	 * @since  2.5
	 */
	private $productEdition = 'full';

	/**
	 * Handle adding credentials to package download request
	 *
	 * @param string $url     url from which package is going to be downloaded
	 * @param array  $headers headers to be sent along the download request (key => value format)
	 *
	 * @return  boolean    true        always true
	 *
	 * @since   2.5
	 * @throws Exception
	 */
	public function onInstallerBeforePackageDownload(&$url, &$headers)
	{
		// are we trying to update our extension?
		foreach ($this->baseUrl as $baseUrl => $productId)
		{
			if (Wb\startsWith($url, $baseUrl))
			{
				$this->productId = $productId;
				break;
			}
		}

		// not one of our URLs.
		if (empty($this->productId))
		{
			return true;
		}

		// read credentials from extension params or any other source
		$credentials = $this->fetchCredentials($url, $headers);

		// bind credentials to request, either in the urls, or using headers
		// or a combination of both
		$this->bindCredentials($credentials, $url, $headers);

		return true;
	}

	/**
	 * Bind credentials to the download request.
	 * In Joomla 4, update key is manager by Joomla and added to the URL.
	 * Prior to J4, we are using a header. Better but not how it was done in J4.
	 *
	 * @param array  $credentials whatever credentials were retrieved for the current user/website
	 * @param string $url         url from which package is going to be downloaded
	 * @param array  $headers     headers to be sent along the download request (key => value format)
	 *
	 * @return void
	 */
	private function bindCredentials($credentials, &$url, &$headers)
	{
		$headers['X-download-auth-ts'] = time();
		$headers['X-download-auth-id'] = $credentials['id'];
	}

	/**
	 * Retrieve user credentials
	 *
	 * @param $url
	 * @param $headers
	 *
	 * @return mixed an array with credentials (id, secret), or null if none found
	 * @throws Exception
	 */
	private function fetchCredentials($url, $headers)
	{
		if (version_compare(JVERSION, '4', 'ge'))
		{
			// j4 adds credentials itself, from update keys manager
			return ['id' => ''];
		}

		// maybe the key was put in URL already by something like Watchful?
		$parsedUrl = parse_url($url, PHP_URL_QUERY);
		if (!empty($parsedUrl))
		{
			$parsedQueryVars = [];
			parse_str($parsedUrl, $parsedQueryVars);
			$urlKey = Wb\arrayGet($parsedQueryVars, 'wblr_k', '');
			if (!empty($urlKey))
			{
				$credentials = ['id' => $urlKey];
			}
		}

		if (empty($credentials['id']))
		{
			// fetch credentials from extension parameters
			// Get the component information from the #__extensions table
			$config      = wblFactory::get()
									 ->getThis(
										 'forseo.config',
										 'system'
									 );
			$credentials = [
				'id' => $config->get('dlid', '')
			];
		}

		// no update id found, display error
		if (empty($credentials['id']))
		{
			$this->loadLanguage();
			$app = Factory::getApplication();
			$app->enqueueMessage(
				Text::sprintf(
					'PLG_INSTALLER_FORSEO_UPDATE_NO_CREDENTIALS',
					'https://weeblr.com/doc/products.forseo/current/getting-started/installation-update/'
				),
				'error'
			);
			$app->redirect('index.php?option=com_installer&view=update');
		}

		return $credentials;
	}
}

