<?php
/**
 * Project: 4SEO
 *
 * @package          4SEO
 * @copyright        Copyright Weeblr llc - 2020-2023
 * @author           Yannick Gaultier - Weeblr llc
 * @license          GNU General Public License version 3; see LICENSE.md
 * @version          4.5.1.2042
 * @date        2023-02-15
 */

namespace Weeblr\Forseo\Model\Extensions;

use Weeblr\Forseo\Data;
use Weeblr\Forseo\Helper;
use Weeblr\Forseo\Model;

use Joomla\CMS\HTML\Registry;
use Joomla\CMS\Component\ComponentHelper;

use Weeblr\Wblib\Forseo\Base;
use Weeblr\Wblib\Forseo\Wb;
use Weeblr\Wblib\Forseo\System;

// Security check to ensure this file is being included by a parent file.
defined('_JEXEC') || defined('WBLIB_EXEC') || die;

class Forsef extends Base\Base
{
	/**
	 * @var System\Logger Convenience logger instance.
	 */
	private $logger = null;

	/**
	 * Instantiate a page object to store current page request data.
	 */
	public function __construct()
	{
		parent::__construct();

		$this->logger = $this->factory->getThe('forseo.logger');
	}

	/**
	 * Hook handler ran when a 4SEF URL pair is manually customized.
	 *
	 * @param array  $data
	 * @param string $originalBasePath
	 * @param array  $customizedSefs
	 *
	 * @return void
	 * @throws \Throwable
	 */
	public function onUrlCustomized($data, $originalBasePath, $customizedSefs)
	{
		$this->logger->debug(__METHOD__ . ', 4SEO URL customized handler, from ' . $originalBasePath . " to \n" . print_r($data, true));

		// prepare
		$targetBasePath = Wb\arrayGet(
			$data,
			'base_path'
		);

		if ($targetBasePath === $originalBasePath)
		{
			$this->logger->debug(__METHOD__ . ', 4SEO URL customized handler: original SEF and customized one are the same, cannot add a redirect for ' . print_r($originalBasePath, true));

			return;
		}

		$db = $this->factory->getThe('db');

		try
		{
			foreach ($customizedSefs as $customizedSef)
			{
				$this->logger->debug(__METHOD__ . ' Customized SEF: ' . $customizedSef);
				$indexableSource = $db->storageSafe($customizedSef);
				$targetSef       = preg_replace(
					'~^' . preg_quote($originalBasePath) . '~',
					$targetBasePath,
					$customizedSef
				);
				$this->logger->debug(__METHOD__ . ' Target SEF: ' . $targetSef);

				$indexableTargetSef = $db->storageSafe($targetSef);

				// update canonical includes
				$db->update(
					'#__forseo_canonical_includes',
					[
						'url' => $indexableTargetSef,
					],
					[
						'url' => $indexableSource
					]
				);

				// update collected URLs
				// NB: we do not attempt to update all the referrers links. This would require to run a full text search
				// of the referrers column to identify rows that have the modified URLs as a referrer, and then update
				// each of them. This can lead to gigantic volumes of db writes. Only way out is to reset analysis and re-crawl.
				$db->update(
					'#__forseo_collected_urls',
					[
						'url'      => $indexableTargetSef,
						'full_url' => $targetSef
					],
					[
						'url' => $indexableSource
					]
				);

				// update custom meta
				$db->update(
					'#__forseo_custom_meta',
					[
						'url' => $indexableTargetSef,
					],
					[
						'url' => $indexableSource
					]
				);

				// update excluded URLs
				$db->update(
					'#__forseo_excluded_urls',
					[
						'url'      => $indexableTargetSef,
						'full_url' => $targetSef
					],
					[
						'url' => $indexableSource
					]
				);

				// update images
				$db->update(
					'#__forseo_images',
					[
						'page_url'      => $indexableTargetSef,
						'page_full_url' => $targetSef
					],
					[
						'page_url' => $indexableSource
					]
				);

				// update links: not updated as we do not know if these links (which are links
				// which triggers a redirect) are the result of a SEF URL action.
				// That's true of pages one could say, but more likely with links.

				// update pages
				$db->update(
					'#__forseo_pages',
					[
						'url'      => $indexableTargetSef,
						'full_url' => $targetSef
					],
					[
						'url' => $indexableSource
					]
				);

				// update perf_data
				$db->update(
					'#__forseo_perf_data',
					[
						'url'      => $indexableTargetSef,
						'full_url' => $targetSef
					],
					[
						'url' => $indexableSource
					]
				);

				// update perf_data_agg
				$db->update(
					'#__forseo_perf_data_agg',
					[
						'url'      => $indexableTargetSef,
						'full_url' => $targetSef
					],
					[
						'url' => $indexableSource
					]
				);

				// update referrers
				$db->update(
					'#__forseo_referrers',
					[
						'url'      => $indexableTargetSef,
						'full_url' => $targetSef
					],
					[
						'url' => $indexableSource
					]
				);

				// update sitemaps_includes
				$db->update(
					'#__forseo_sitemaps_includes',
					[
						'url' => $indexableTargetSef,
					],
					[
						'url' => $indexableSource
					]
				);

			}
		}
		catch (\Throwable $e)
		{
			$this->logger->error(__METHOD__ . ' %s::%d %s - %s', $e->getFile(), $e->getLine(), $e->getMessage(), $e->getTraceAsString());
		}
	}
}
