<?php
/**
 * Project: 4SEO
 *
 * @package          4SEO
 * @copyright        Copyright Weeblr llc - 2020-2023
 * @author           Yannick Gaultier - Weeblr llc
 * @license          GNU General Public License version 3; see LICENSE.md
 * @version          4.5.1.2042
 * @date        2023-02-15
 */

namespace Weeblr\Forseo\View;

use Weeblr\Forseo\Data;
use Weeblr\Forseo\Helper;

use Weeblr\Wblib\Forseo\Wb;
use Weeblr\Wblib\Forseo\System;
use Weeblr\Wblib\Forseo\Mvc;
use Weeblr\Wblib\Forseo\Html;
use Weeblr\Wblib\Forseo\Joomla\StringHelper\StringHelper;


/** ensure this file is being included by a parent file */
defined('_JEXEC') || defined('WBLIB_EXEC') || die;

/**
 * Main admin view. Loads up custom CSS/Js per platform then the app itself.
 */
class Admin extends Mvc\ViewHtml
{
	/**
	 * @var System\Hook
	 */
	protected $hooks;

	/**
	 * @var Html\Assetsmanager
	 */
	protected $assetsManager;

	/**
	 * @var bool
	 */
	protected $isDev;

	/**
	 * Constructor
	 *
	 * @param array $options An array of options.
	 */
	public function __construct($options = array())
	{

		parent::__construct($options);

		$this->hooks         = $this->factory->getThe('hook');
		$this->isDev         = 'dev' === WBLIB_Forseo_OP_MODE;
		$this->assetsManager = $this->factory->getThe('forseo.assetsManager');
	}

	/**
	 * Renders the view content, returning it in a string and
	 * optionally echoing it
	 */
	protected function doRender()
	{
		try
		{
			/**
			 * Filter custom CSS for admin.
			 *
			 * @api     forseo
			 * @package 4SEO\filter\admin
			 * @var forseo_admin_ui_constants
			 * @since   1.0.0
			 *
			 * @param array $uiConstants List of constants related to the specific platform visual display.
			 *
			 * @return array
			 *
			 */
			$uiConstants = $this->hooks->filter(
				'forseo_admin_ui_constants',
				[]
			);

			$defaultStyles = '';
			/**
			 * Filter custom CSS for admin.
			 *
			 * @param string $defaultStyles The raw css to be inserted as style tag in the page.
			 *
			 * @return string
			 * @var forseo_custom_admin_css
			 *
			 * @api     forseo
			 *
			 * @package 4SEO\filter\admin
			 *
			 * @since   1.0.0
			 *
			 */
			$css = $this->hooks->filter(
				'forseo_custom_admin_css',
				$defaultStyles
			);

			// Get and filter the javascript
			/**
			 * Filter main JS bundle URL.
			 *
			 * @param string $url The url of the js bundle to be linked from the page.
			 *
			 * @return string
			 * @var forseo_admin_js_bundle_url
			 *
			 * @api     forseo
			 *
			 * @package 4SEO\filter\bundle
			 *
			 * @since   1.0.0
			 *
			 */
			$forseoAppJs = $this->hooks->filter(
				'forseo_admin_js_bundle_url',
				$this->isDev
					? FORSEO_APP_DEV_JS_BUNDLE_ADMIN
					: $this->assetsManager->getHashedMediaLink('admin.js')
			);

			/**
			 * Filter features overrides
			 *
			 * @param array $featuresOverrides
			 *
			 * @return arrat
			 * @var forseo_features_overrides
			 *
			 * @api     forseo
			 *
			 * @package 4SEO\filter\features
			 *
			 * @since   4.1.2
			 *
			 */
			$featuresOverrides = $this->hooks->filter(
				'forseo_features_overrides',
				FORSEO_FEATURES_OVERRIDES
			);

			// Load language strings
			$currentLanguageTag = $this->platform->getCurrentLanguageTag();
			$languageStrings    = $this->factory
				->getA(Helper\Language::class)
				->getJsLanguageStrings(
					['common', 'admin'],
					$currentLanguageTag
				);

			// Fetch known search engines updates
			$searchEnginesUpdates = $this->factory->getThe('forseo.searchEnginesHelper')
												  ->getEnginesUpdates();

			// Hardcode in the HTML configuration for the app, including language strings
			$user            = $this->platform->getUser();
			$rewritingPrefix = $this->platform->getUrlRewritingPrefix();
			$rewritingPrefix = empty($rewritingPrefix)
				? $rewritingPrefix
				: StringHelper::ltrim($rewritingPrefix, '/') . '/';

			$systemConfig   = $this->factory->getThis('forseo.config', 'system');
			$sh404sefConfig = $this->factory->getThis('forseo.config', 'sh404sef');
			$pagesConfig    = $this->factory->getThis('forseo.config', 'pages');
			$sdConfig       = $this->factory->getThis('forseo.config', 'sd');
			$insightsConfig = $this->factory->getThis('forseo.config', 'insights');
			$forSeoJsConfig = [
				'platformVersion'              => [
					'major'   => $this->platform->majorVersion(),
					'version' => $this->platform->version()
				],
				'version'                      => [
					'current' => '4.5.1.2042',
					'update'  => $this->factory->getA(Helper\Update::class)->latestVersion(),
					'edition' => 'full',
					'api'     => 'v1'
				],
				'tz'                           => $this->platform->getTimeZone(),
				'date'                         => '2023-02-15',
				'copyright'                    => 'Copyright Weeblr llc - 2020-2023',
				'language'                     => $currentLanguageTag,
				'languages'                    => $this->platform->getFrontendLanguages(),
				'isDev'                        => $this->isDev,
				'isMultilingual'               => $this->platform->isMultilingual(),
				'configWizardCompleted'        => $systemConfig->isTruthy('configWizardCompleted'),
				'canImportMetaFromSh404sef'    => $sh404sefConfig->get('canImportMetaFromSh404sef', -1),
				'canImportAliasesFromSh404sef' => $sh404sefConfig->get('canImportAliasesFromSh404sef', -1),
				'importWizardCompleted'        => $sh404sefConfig->isTruthy('importWizardCompleted'),
				'importWizardRunMode'          => $sh404sefConfig->get('importWizardRunMode'),
				'canAutoClearNotifications'    => $systemConfig->isTruthy('canAutoClearNotifications'),
				'displayConfig'                => [],
				'urls'                         => [
					'root'             => $this->platform->getRootUrl(true),
					'rootFull'         => $this->platform->getRootUrl(false),
					'canonicalRoot'    => $pagesConfig->get('canonicalRootUrl'),
					'rewritePrefix'    => $rewritingPrefix,
					'base'             => $this->platform->getBaseUrl(true),
					'baseFull'         => $this->platform->getBaseUrl(false),
					'xmlSitemap'       => $this->factory->getA(Helper\Sitemaps::class)->xmlUrl(),
					'images'           => $this->platform->getRootUrl(false) . 'images',
					'api'              => System\Route::normalizePath(
						$this->platform->getBaseUrl(true)
						. '/index.php?'
						. $this->factory->getThe('api')->getSlug()
						. '=/forseo'),
					'helpRequirements' => 'https://weeblr.com/doc/requirements',
					'poauth'           => $this->factory->getThis('forseo.config', 'integrations')->get('oAuthProxyEndpoint'),
				],
				'user'                         => [
					'id' => empty($user->id) ? 'default' : substr(md5($user->id), 0, 12)
				],
				'path'                         => [
					'images' => System\Route::normalizePath(FORSEO_APP_ASSETS_BASE_URL . '/images'),
					'js'     => System\Route::normalizePath(FORSEO_APP_ASSETS_BASE_URL . '/js'),
					'editor' => '/tmce/tinymce.min.4.9.11.js'
				],
				'tokens'                       => [
					'csrf' => $this->platform->getCSRFToken()
				],
				'bundleFile'                   => $forseoAppJs,
				'uiConstants'                  => $uiConstants,
				'sharingImageSpec'             => $this->factory->getThis('forseo.config', 'socialnetworks')->get('imageSharingSpec'),
				'imageSpec'                    => $sdConfig->get('imageSpec'),
				'logoSpec'                     => $sdConfig->get('logoSpec'),
				'eventImageSpec'               => $sdConfig->get('eventImageSpec'),
				'sdLocalBusinessTypes'         => Data\Sd::LOCAL_BUSINESS_TYPES,
				'countriesAlpha2'              => System\Data::COUNTRIES_ISO3166_1_ALPHA_2,
				'currencies4217'               => System\Data::CURRENCIES_ISO_4217,
				'defaultGoogleMapsApiKey'      => $this->factory->getThis('forseo.config', 'app')->get('apiKeys.googleMaps.1'),
				'perfThresholds'               => $pagesConfig->get('perfThresholds'),
				'gscThresholds'                => $insightsConfig->get('gscThresholds'),
				'featuresOverrides'            => $featuresOverrides,
				'searchEnginesUpdates'         => $searchEnginesUpdates,
				'siteName'                     => $this->platform->getSitename()
			];

			$jsBundles  = $this->getJsBundles($forseoAppJs);
			$cssBundles = $this->getCssBundles();


			// finally insert into page
			/**
			 * Filter custom JS for admin.
			 *
			 * @param string $js The javascript to be inserted into the page with a script tag.
			 *
			 * @package 4SEO\filter\admin
			 *
			 * @var forseo_custom_admin_js
			 *
			 * @api     forseo
			 *
			 * @since   1.0.0
			 *
			 */
			$customJs = $this->hooks->filter(
				'forseo_custom_admin_js',
				''
			);

			$js = "\nvar forSeoConfig = " . System\Strings::jsonPrettyPrint($forSeoJsConfig) . ";\n";
			$js .= "\nvar forSeoLanguageStrings = " . System\Strings::jsonPrettyPrint($languageStrings) . ";\n";
			$js .= "\n" . $customJs;

			$this->platform
				->addScripts($jsBundles)
				->addStylesheets($cssBundles)
				->addStyleDeclaration($css)
				->addScriptDeclaration($js);

			$this->platform->setAdminTitle('4SEO - All Joomla! SEO');

			return parent::doRender();
		}
		catch (\Throwable $e)
		{
			$this->factory->getThe('forseo.logger')->error('%s::%d %s - %s', $e->getFile(), $e->getLine(), $e->getMessage(), $e->getTraceAsString());
		}
	}


	/**
	 * Builds and filters definition array of javascript bundles to be included in page.
	 *
	 * @return array
	 */
	private function getJsBundles($appJs)
	{
		/**
		 * Filter list of js files to be inserted before main bundle.
		 *
		 * Each record added to the lsit must be:
		 * [
		 *   'url' => '', mandatory
		 *   'options' => [], | []
		 *   'attr' => [], | []
		 * ]
		 *
		 * @param array $jsDefs An array of array[url, options, attr], each defining a link to a js file.
		 *
		 * @return string
		 * @var forseo_custom_admin_js
		 *
		 * @api     forseo
		 *
		 * @package 4SEO\filter\bundle
		 *
		 * @since   1.0.0
		 *
		 */
		$jsBundlesBefore = $this->hooks->filter(
			'forseo_js_before_bundle',
			[]
		);

		/**
		 * Filter list of js files to be inserted after main bundle.
		 *
		 * Each record added to the lsit must be:
		 * [
		 *   'url' => '', mandatory
		 *   'options' => [], | []
		 *   'attr' => [], | []
		 * ]
		 *
		 * @param array $jsDefs An array of array[url, options, attr], each defining a link to a js file.
		 *
		 * @return string
		 * @var forseo_custom_admin_js
		 *
		 * @api     forseo
		 *
		 * @package 4SEO\filter\bundle
		 *
		 * @since   1.0.0
		 *
		 */
		$jsBundlesAfter = $this->hooks->filter(
			'forseo_js_after_bundle',
			[]
		);

		return array_merge(
			$jsBundlesBefore,
			[
				[
					'url'     => $appJs,
					'options' => [],
					'attr'    => $this->isDev
						? ['type' => 'module']
						: ['defer' => 'defer']
				]
			],
			$jsBundlesAfter
		);
	}

	/**
	 * Builds and filters definition array of css bundles to be included in page.
	 *
	 * @return array
	 */
	private function getCssBundles()
	{
		/**
		 * Filter list of css files to be inserted before main bundle.
		 *
		 * Each record added to the lsit must be:
		 * [
		 *   'url' => '', mandatory
		 *   'options' => [], | []
		 *   'attr' => [], | []
		 * ]
		 *
		 * @param array $cssDefs An array of array[url, options, attr], each defining a link to a css file.
		 *
		 * @return string
		 * @var forseo_custom_admin_js
		 *
		 * @api     forseo
		 *
		 * @package 4SEO\filter\bundle
		 *
		 * @since   1.0.0
		 *
		 */
		$cssBundlesBefore = $this->hooks->filter(
			'forseo_css_before_bundle',
			[]
		);

		/**
		 * Filter list of css files to be inserted after main bundle.
		 *
		 * Each record added to the lsit must be:
		 * [
		 *   'url' => '', mandatory
		 *   'options' => [], | []
		 *   'attr' => [], | []
		 * ]
		 *
		 * @param array $cssDefs An array of array[url, options, attr], each defining a link to a css file.
		 *
		 * @return string
		 * @var forseo_custom_admin_js
		 *
		 * @api     forseo
		 *
		 * @package 4SEO\filter\bundle
		 *
		 * @since   1.0.0
		 *
		 */
		$cssBundlesAfter = $this->hooks->filter(
			'forseo_css_after_bundle',
			[]
		);

		return array_merge(
			$cssBundlesBefore,
			[
				[
					'url' => $this->assetsManager->getHashedMediaLink('admin.base.css')
				],
				[
					'url' => $this->assetsManager->getHashedMediaLink('admin.css')
				]
			],
			$cssBundlesAfter
		);
	}

}